@@ -1,5 +1,5 @@
 ;; -*- no-byte-compile: t; lexical-binding: nil -*-
-(define-package "justl" "0.14"
+(define-package "justl" "20250909.422"
   "Major mode for driving just files."
   '((transient  "0.1.0")
     (emacs      "27.1")
@@ -7,6 +7,6 @@
     (f          "0.20.0")
     (inheritenv "0.2"))
   :url "https://github.com/psibi/justl.el"
-  :commit "1b16ca44b227dc761a2c8631690140f62e024f98"
-  :revdesc "v0.14-0-g1b16ca44b227"
+  :commit "8096ebd602f374936be2c14cbc6b2e7768c3fccf"
+  :revdesc "8096ebd602f3"
   :keywords '("just" "justfile" "tools" "processes"))
