@@ -19,8 +19,8 @@
 ;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 ;; USA
 
-;; Package-Version: 0.14
-;; Package-Revision: v0.14-0-g1b16ca44b227
+;; Package-Version: 20250909.422
+;; Package-Revision: 8096ebd602f3
 ;; Author: Sibi Prabakaran
 ;; Keywords: just justfile tools processes
 ;; URL: https://github.com/psibi/justl.el
@@ -52,9 +52,9 @@
 ;; ? => help popup
 ;; g => refresh
 ;; e => execute recipe
-;; E => execute recipe with eshell
+;; E => execute recipe with a shell
 ;; w => execute recipe with arguments
-;; W => open eshell without executing
+;; W => open a shell without executing
 ;;
 ;; Customize:
 ;;
@@ -64,6 +64,10 @@
 ;; You can also control the width of the RECIPE column in the justl
 ;; buffer via `justl-recipe width`.  By default it has a value of 20.
 ;;
+;; You can change the shell between `eshell' and `vterm' using the
+;; `justl-shell' variable.  Using vterm requires the `vterm' package to
+;; be installed.
+;;
 
 ;;; Code:
 
@@ -90,10 +94,12 @@
 (defcustom justl-executable "just"
   "Location of just executable."
   :type 'file
+  :group 'justl
   :safe 'stringp)
 
 (defcustom justl-recipe-width 20
   "Width of the recipe column."
+  :group 'justl
   :type 'integer)
 
 (defcustom justl-justfile nil
@@ -102,17 +108,37 @@
 If this is NIL, it means that no justfile was found.  In any
 other cases, it's a known path."
   :type 'string
+  :group 'justl
   :local t
   :safe 'stringp)
 
-(defcustom justl-include-private-recipes t
+(defcustom justl-include-private-recipes nil
   "If non-nil, include private recipes in the list."
   :type 'boolean
+  :group 'justl
   :safe 'booleanp)
 
 (defcustom justl-per-recipe-buffer nil
   "If non-nil, create a new buffer per recipe."
   :type 'boolean
+  :group 'justl
+  :safe 'booleanp)
+
+(defcustom justl-shell 'eshell
+  "Shell to use when running recipes.
+Can be either Eshell, vterm, or eat.  Using vterm requires the vterm
+package to be installed.  Using eat requires the eat package to be installed."
+
+  :type '(choice (const eshell)
+                 (const vterm)
+                 (const eat))
+  :group 'justl)
+
+(defcustom justl-pop-to-buffer-on-display t
+  "If non-nil, selects the justl output buffer when it is displayed.
+If nil, displays the output buffer without selecting it."
+  :type 'boolean
+  :group 'justl
   :safe 'booleanp)
 
 (defun justl--recipe-output-buffer (recipe-name)
@@ -154,7 +180,7 @@ NAME is the buffer name."
 Returns the absolute path if file exists or nil if no path
 was found."
   (cl-flet*
-      ((is-justfile (s) (string= "justfile" (downcase s)))
+      ((is-justfile (s) (s-ends-with? "justfile" s t))
        (any-justfile (d) (seq-find #'is-justfile (directory-files d))))
     (when-let ((location (locate-dominating-file dir #'any-justfile)))
       (expand-file-name (any-justfile location) location))))
@@ -164,7 +190,7 @@ was found."
 
 DIR represents the directory where search will be carried out.
 It searches either for the filename justfile or .justfile"
-  (when-let ((justfile-path (justl--find-any-justfile dir)))
+  (when-let ((justfile-path (or (justl--find-any-justfile dir) (getenv "JUST_JUSTFILE"))))
     (setq-local justl-justfile justfile-path)
     justfile-path))
 
@@ -233,7 +259,7 @@ controls if we are going to display the process status on mode line."
               (eq compilation-scroll-output 'first-error))
           (set (make-local-variable 'compilation-auto-jump-to-next) t)))))
 
-(defvar justl--compile-command nil
+(defvar-local justl--compile-command nil
   "Last shell command used to do a compilation; default for next compilation.")
 
 (defun justl--make-process (command &optional args)
@@ -255,7 +281,8 @@ ARGS is a plist that affects how the process is run.
     (justl-compilation-setup-buffer buf directory mode)
     (with-current-buffer buf
       (insert (format "Just target execution started at %s \n\n" (substring (current-time-string) 0 19)))
-      (let* ((process (apply
+      (let* ((default-directory directory)
+	     (process (apply
                        #'start-file-process process-name buf command)))
         (setq justl--compile-command command)
         (setq-local justl-justfile (justl--justfile-from-arg (elt command 1)))
@@ -263,7 +290,9 @@ ARGS is a plist that affects how the process is run.
         (set-process-filter process 'justl--process-filter)
         (set-process-sentinel process (lambda (proc _) (justl--sentinel proc buf)))
         (set-process-coding-system process 'utf-8-emacs-unix 'utf-8-emacs-unix)
-        (pop-to-buffer buf)))))
+        (if justl-pop-to-buffer-on-display
+            (pop-to-buffer buf)
+          (display-buffer buf))))))
 
 (defvar justl-compile-mode-map
   (let ((map (make-sparse-keymap)))
@@ -273,10 +302,31 @@ ARGS is a plist that affects how the process is run.
     map)
   "Keymap for justl compilation log buffers.")
 
+(defun justl-set-new-working-dir (dir)
+  "Set DIR as the new working directory.
+This is usually used with no-cd recipe attribute."
+  (interactive "DNew Working Directory:")
+  (setq-local default-directory dir))
+
 (defun justl-recompile ()
   "Execute the same just target again."
   (interactive)
-  (justl--make-process justl--compile-command (list :buffer justl--output-process-buffer
+  ;; This is copied and adapted from `compilation-start'.
+  (let ((comp-proc (get-buffer-process (current-buffer))))
+    (if comp-proc
+        (if (or (not (eq (process-status comp-proc) 'run))
+                (eq (process-query-on-exit-flag comp-proc) nil)
+                (yes-or-no-p "The last target is still running; kill it? "))
+            (condition-case ()
+                (progn
+                  (interrupt-process comp-proc)
+                  (sit-for 1)
+                  (delete-process comp-proc))
+              (error nil))
+          (error "Cannot have two processes in `%s' at once"
+                 (buffer-name)))))
+
+  (justl--make-process justl--compile-command (list :buffer (buffer-name)
                                                     :process "just"
                                                     :directory (if justl-justfile
                                                                    (f-dirname justl-justfile)
@@ -312,7 +362,7 @@ Error matching regexes from compile.el are removed."
 
 PROCESS-NAME is an identifier for the process.  Default to \"just\".
 RECIPE-NAME is the name of the recipe.
-ARGS is a ist of arguments."
+ARGS is a list of arguments."
   (when (equal process-name "")
     (setq process-name justl-executable))
   (let ((buffer-name (justl--recipe-output-buffer recipe-name))
@@ -326,17 +376,18 @@ ARGS is a ist of arguments."
     (justl--log-command process-name cmd)
     (justl--make-process cmd (list :buffer buffer-name
                                    :process process-name
+				   :directory default-directory
                                    :mode mode))))
 
-(defun justl--exec-without-justfile (process-name args)
+(defun justl--exec-without-justfile (process-name args buffer-name)
   "Utility function to run commands in the proper setting.
 
 PROCESS-NAME is an identifier for the process.  Default to \"just\".
-ARGS is a ist of arguments."
+ARGS is a ist of arguments.
+BUFFER-NAME is the output buffer name."
   (when (equal process-name "")
     (setq process-name justl-executable))
-  (let ((buffer-name justl--output-process-buffer)
-        (error-buffer (justl--process-error-buffer process-name))
+  (let ((error-buffer (justl--process-error-buffer process-name))
         (cmd (append (list justl-executable) args))
         (mode 'justl-compile-mode))
     (when (get-buffer buffer-name)
@@ -348,7 +399,9 @@ ARGS is a ist of arguments."
                                    :process process-name
                                    :directory default-directory
                                    :mode mode)))
-  (pop-to-buffer justl--output-process-buffer))
+  (if justl-pop-to-buffer-on-display
+      (pop-to-buffer buffer-name)
+    (display-buffer buffer-name)))
 
 (defun justl--exec-to-string-with-exit-code (executable &rest args)
   "Run EXECUTABLE with ARGS, throwing an error if the command fails.
@@ -388,27 +441,52 @@ Logs the command run."
                             (cdr recipes-entry)))
           parsed)))))
 
+(cl-defstruct recipe
+  ;; Recipe name
+  name
+  ;; Optional recipe documentation
+  doc
+  ;; Parameters for the recipe
+  parameters
+  ;; Is the recipe private
+  private)
+
 (defun justl--get-recipes (justfile)
   "Return all the recipes from JUSTFILE.
 They are returned as objects, as per the JSON output of \"just --dump\"."
   (let-alist (justl--parse justfile)
-    (mapcar 'cdr .recipes)))
+    (let ((all-recipes (mapcar (lambda (x) (make-recipe :name (alist-get 'name x)
+				     :doc (alist-get 'doc x)
+				     :parameters (alist-get 'parameters x)
+				     :private (alist-get 'private x))) .recipes)))
+      (if justl-include-private-recipes
+	  all-recipes
+	(seq-filter (lambda (recipe) (not (justl--recipe-private-p recipe))) all-recipes)))))
+
+;;; todo: For easily integrating it, we need something like this
+;;; integrated upstream:
+;;; https://github.com/casey/just/issues/2252#issuecomment-2474171211
+(defun justl--get-modules (justfile)
+  "Return all the modules from JUSTFILE.
+They are returned as objects, as per the JSON output of \"just --dump\"."
+  (let-alist (justl--parse justfile)
+    (mapcar 'car .modules)))
 
 (defun justl--recipe-name (recipe)
   "Get the name of RECIPE."
-  (let-alist recipe .name))
+  (recipe-name recipe))
 
 (defun justl--recipe-desc (recipe)
   "Get the description of RECIPE."
-  (let-alist recipe .doc))
+  (recipe-doc recipe))
 
 (defun justl--recipe-args (recipe)
   "Get the arguments for RECIPE."
-  (let-alist recipe .parameters))
+  (recipe-parameters recipe))
 
 (defun justl--recipe-private-p (recipe)
   "Return non-nil if RECIPE is private."
-  (let-alist recipe (cl-find "private" .attributes :test 'string=)))
+  (recipe-private recipe))
 
 (defun justl--arg-name (arg)
   "Get the name of argument ARG."
@@ -427,6 +505,7 @@ They are returned as objects, as per the JSON output of \"just --dump\"."
   (when arg
     (cadr (s-split "--justfile=" arg))))
 
+;;;###autoload
 (defun justl-exec-recipe-in-dir ()
   "Populate and execute the selected recipe."
   (interactive)
@@ -437,36 +516,38 @@ They are returned as objects, as per the JSON output of \"just --dump\"."
            (recipe-name (completing-read "Recipes: "
                                          (mapcar 'justl--recipe-name recipes)
                                          nil t nil nil))
-           (recipe (cdr (assoc recipe-name recipes))))
+           (recipe (justl--find-recipes recipes recipe-name)))
       (justl--exec-without-justfile
        justl-executable
        (cons recipe-name
              (mapcar 'justl--read-arg
-                     (justl--recipe-args recipe)))))))
+                     (justl--recipe-args recipe)))
+       (justl--recipe-output-buffer recipe-name)))))
 
+;;;###autoload
 (defun justl-exec-default-recipe ()
   "Execute default recipe."
   (interactive)
-  (justl--exec-without-justfile justl-executable nil))
+  (justl--exec-without-justfile justl-executable nil (justl--recipe-output-buffer "default")))
 
 (defvar justl-mode-map
   (let ((map (make-sparse-keymap)))
     (define-key map (kbd "g") 'justl--refresh-buffer)
     (define-key map (kbd "e") 'justl-exec-recipe)
-    (define-key map (kbd "E") 'justl-exec-eshell)
+    (define-key map (kbd "E") 'justl-exec-shell)
     (define-key map (kbd "?") 'justl-help-popup)
     (define-key map (kbd "h") 'justl-help-popup)
     (define-key map (kbd "w") 'justl--exec-recipe-with-args)
-    (define-key map (kbd "W") 'justl-no-exec-eshell)
+    (define-key map (kbd "W") 'justl-no-exec-shell)
+    (define-key map (kbd "m")  'justl--show-modules)
     (define-key map (kbd "RET") 'justl-go-to-recipe)
     map)
   "Keymap for `justl-mode'.")
 
-(defun justl--buffer-name ()
+(defun justl--buffer-name (is-module)
   "Return justl buffer name."
   (let ((justfile (justl--find-justfile default-directory)))
-    (format "*just [%s]*"
-            (f-dirname justfile))))
+    (format "*just [%s] %s*" justfile (if is-module "(Module)" ""))))
 
 (defun justl--tabulated-entries (recipes)
   "Turn RECIPES to tabulated entries."
@@ -476,10 +557,17 @@ They are returned as objects, as per the JSON output of \"just --dump\"."
              (justl--recipe-name r)
              (vector (propertize (justl--recipe-name r) 'recipe r)
                      (or (justl--recipe-desc r) ""))))
-          (seq-filter (lambda (r)
-                        (or justl-include-private-recipes
-                            (not (justl--recipe-private-p r))))
-                      recipes)))
+          recipes))
+
+(defun justl--tabulated-modules-entries (modules)
+  "Turn MODULES to tabulated entries."
+  (mapcar (lambda (m)
+            (list
+             ;; Comparison key, used to restore line number after refresh
+             (just-module-name m)
+             (vector (propertize (just-module-name m) 'module m)
+                     (or (just-module-doc m) ""))))
+          modules))
 
 (defun justl-exec-eshell (&optional no-send)
   "Execute just recipe in eshell.
@@ -507,6 +595,80 @@ not executed."
   (interactive)
   (justl-exec-eshell t))
 
+(defun justl-exec-vterm (&optional no-send)
+  "Execute just recipe in vterm.
+When NO-SEND is non-nil, the command is inserted ready for editing but
+is not executed."
+  (interactive)
+  (unless (require 'vterm nil t)
+    (user-error "Package `vterm' was not found!"))
+  (let* ((recipe (justl--get-recipe-under-cursor))
+         (vterm-buffer-name (format "justl - vterm - %s" (justl--recipe-name recipe)))
+         (default-directory (f-dirname justl-justfile)))
+    (vterm)
+
+    (let* ((recipe-name (justl--recipe-name recipe))
+           (recipe-args (justl--recipe-args recipe))
+           (transient-args (transient-args 'justl-help-popup))
+           (args-list (append (list "exec" justl-executable)
+                              transient-args
+                              (list recipe-name)
+                              (mapcar 'justl--arg-default recipe-args))))
+      (vterm-insert (string-join args-list " ")))
+    (unless no-send
+      (vterm-send-return))))
+
+(defun justl-no-exec-vterm ()
+  "Open vterm with the recipe but do not execute it."
+  (interactive)
+  (justl-exec-vterm t))
+
+(defun justl-exec-eat (&optional no-send)
+  "Execute just recipe in eat.
+When NO-SEND is non-nil, the command is inserted ready for editing but
+is not executed."
+  (interactive)
+  (unless (require 'eat nil t)
+    (user-error "Package `eat' was not found!"))
+  (let* ((recipe (justl--get-recipe-under-cursor))
+         (eat-buffer-name (format "justl - eat - %s" (justl--recipe-name recipe)))
+         (default-directory (f-dirname justl-justfile)))
+    (eat)
+    (let ((eat-buffer (eat eat-buffer-name)))
+      (with-current-buffer eat-buffer
+        (let* ((recipe-name (justl--recipe-name recipe))
+               (recipe-args (justl--recipe-args recipe))
+               (transient-args (transient-args 'justl-help-popup))
+               (args-list (cons justl-executable
+                                (append transient-args
+                                        (list recipe-name)
+                                        (mapcar 'justl--arg-default recipe-args)))))
+          (let ((command-string (string-join args-list " ")))
+            (eat-term-send-string eat-terminal command-string))
+          (unless no-send
+            (eat-term-send-string eat-terminal "\r")))))))
+
+(defun justl-no-exec-eat ()
+  "Open eat with the recipe but do not execute it."
+  (interactive)
+  (justl-exec-eat t))
+
+(defun justl-exec-shell (&optional no-send)
+  "Execute just recipe in `justl-shell'.
+When NO-SEND is non-nil, the command is inserted ready for editing but
+is not executed."
+  (interactive)
+  (pcase justl-shell
+    ('eshell (justl-exec-eshell no-send))
+    ('vterm (justl-exec-vterm no-send))
+    ('eat (justl-exec-eat no-send))
+    (_ (user-error "Invalid value for `justl-shell'"))))
+
+(defun justl-no-exec-shell ()
+  "Open `justl-shell' with the recipe but do not execute it."
+  (interactive)
+  (justl-exec-shell t))
+
 (transient-define-argument justl--color ()
   :description "Color output"
   :class 'transient-switches
@@ -532,13 +694,19 @@ not executed."
     ;; global
     ("g" "Refresh" justl)
     ("e" "Exec" justl-exec-recipe)
-    ("E" "Exec with eshell" justl-exec-eshell)
+    ("E" "Exec with shell" justl-exec-shell)
     ("w" "Exec with args" justl--exec-recipe-with-args)
-    ("W" "Open eshell with args" justl-no-exec-eshell)
+    ("W" "Open shell with args" justl-no-exec-shell)
+    ("m" "Show modules" justl--show-modules)
     ("RET" "Go to recipe" justl-go-to-recipe)
     ]
    ])
 
+(defun justl--show-modules ()
+  "Open modules buffer."
+  (interactive)
+  (justl-module))
+
 (defun justl--read-arg (arg)
   "Read a value for ARG from the minibuffer."
   (let ((default (justl--arg-default arg)))
@@ -549,6 +717,10 @@ not executed."
        ('nil nil)
        (_ default)))))
 
+(defun justl--find-recipes (recipes name)
+  "Return recipe from RECIPES matching NAME."
+  (seq-first (seq-filter (lambda (recipe) (string= (recipe-name recipe) name)) recipes)))
+
 (defun justl-exec-recipe ()
   "Execute just recipe."
   (interactive)
@@ -571,11 +743,10 @@ not executed."
                              recipe-name))))
     (justl--exec
      justl-executable
-     recipe-name
+     recipe-name			; Buffer name
      (append (transient-args 'justl-help-popup)
-             (cons
-              recipe-name
-              (split-string user-args " "))))))
+	     (list recipe-name)
+	     (split-string user-args " ")))))
 
 (defun justl-go-to-recipe ()
   "Go to the recipe on justfile."
@@ -593,14 +764,49 @@ not executed."
         (get-text-property 0 'recipe (aref entry 0))
       (user-error "There is no recipe on the current line"))))
 
+(defun justl--get-module-under-cursor ()
+  "Utility function to get the name of the module under the cursor."
+  (let ((entry (tabulated-list-get-entry)))
+    (if entry
+        (get-text-property 0 'module (aref entry 0))
+      (user-error "There is no module on the current line"))))
+
 (defun justl--refresh-buffer ()
   "Refresh justl buffer."
   (interactive)
   (let* ((justfile (justl--find-justfile default-directory))
          (entries (justl--get-recipes justfile)))
     (setq tabulated-list-entries (justl--tabulated-entries entries))
+    (tabulated-list-print t)))
+
+(cl-defstruct just-module
+  ;; Module name
+  name
+  ;; Optional module documentation
+  doc
+  ;; Source path of the module
+  source
+  )
+
+(defun justl--module-refresh-buffer ()
+  "Refresh justl module buffer."
+  (interactive)
+  (let* ((justfile (justl--find-justfile default-directory))
+         (modules (justl--get-modules justfile)))
+    (setq tabulated-list-entries (justl--tabulated-modules-entries modules))
     (tabulated-list-print t)
-    (message "justl-mode: Refreshed")))
+    (message "justl-module-mode: Refreshed")))
+
+;;; todo: For easily integrating it, we need something like this
+;;; integrated upstream:
+;;; https://github.com/casey/just/issues/2252#issuecomment-2474171211
+(defun justl--get-modules (justfile)
+  "Return all the modules from JUSTFILE.
+They are returned as objects, as per the JSON output of \"just --dump\"."
+  (let-alist (justl--parse justfile)
+    (mapcar (lambda (x) (make-just-module :name (symbol-name (car x))
+					  :doc (alist-get 'doc (cdr x))
+					  :source (alist-get 'source (cdr x)))) .modules)))
 
 ;;;###autoload
 (defun justl ()
@@ -608,12 +814,13 @@ not executed."
   (interactive)
   (unless (justl--find-justfile default-directory)
     (error "No justfile found"))
-  (justl--pop-to-buffer (justl--buffer-name))
+  (justl--pop-to-buffer (justl--buffer-name nil))
   (justl-mode)
   (justl--refresh-buffer))
 
 (define-derived-mode justl-mode tabulated-list-mode  "Justl"
   "Special mode for justl buffers."
+  :group 'justl
   (buffer-disable-undo)
   (setq truncate-lines t)
   (setq tabulated-list-format
@@ -624,5 +831,85 @@ not executed."
   (tabulated-list-print t)
   (hl-line-mode 1))
 
+;;;###autoload
+(defun justl-module ()
+  "Invoke the justl-module buffer."
+  (interactive)
+  (unless (justl--find-justfile default-directory)
+    (error "No justfile found"))
+  (justl--pop-to-buffer (justl--buffer-name t))
+  (justl-module-mode)
+  (justl--module-refresh-buffer))
+
+(defvar justl-module-mode-map
+  (let ((map (make-sparse-keymap)))
+    (define-key map (kbd "g") 'justl--module-refresh-buffer)
+    (define-key map (kbd "e") 'justl-exec-module)
+    (define-key map (kbd "?") 'justl-module-help-popup)
+    (define-key map (kbd "h") 'justl-module-help-popup)
+    (define-key map (kbd "o") 'justl--module-open-justl)
+    (define-key map (kbd "RET") 'justl--go-to-module)
+    map)
+  "Keymap for `justl-module-mode'.")
+
+(define-derived-mode justl-module-mode tabulated-list-mode  "Justl (M)"
+  "Special mode for justl module buffers."
+  :group 'justl
+  (buffer-disable-undo)
+  (setq truncate-lines t)
+  (setq tabulated-list-format
+        (vector (list "MODULES" justl-recipe-width t)
+                (list "DESCRIPTION" 20 t)))
+  (setq tabulated-list-sort-key nil)
+  (tabulated-list-init-header)
+  (tabulated-list-print t)
+  (hl-line-mode 1))
+
+(transient-define-prefix justl-module-help-popup ()
+  "Justl Module Menu."
+  [["Arguments"
+    ("-s" "Clear shell arguments" "--clear-shell-args")
+    ("-d" "Dry run" "--dry-run")
+    ("-e" "Disable .env file" "--no-dotenv")
+    ("-h" "Highlight" "--highlight")
+    ("-n" "Disable Highlight" "--no-highlight")
+    ("-q" "Quiet" "--quiet")
+    ("-v" "Verbose output" "--verbose")
+    ("-u" "Unstable" "--unstable")
+    (justl--color)
+    ]
+   ["Actions"
+    ;; global
+    ("g" "Refresh" justl-module)
+    ("e" "Exec" justl-exec-module)
+    ("o" "Open justl buffer for the module" justl--module-open-justl)
+    ("RET" "Go to module" justl--go-to-module)
+    ]
+   ])
+
+(defun justl-exec-module ()
+  "Execute first recipe of the module."
+  (interactive)
+  (let* ((module (justl--get-module-under-cursor))
+	 (default-directory (f-dirname (just-module-source module))))
+    (justl--exec
+     justl-executable
+     (just-module-name module)
+     (append (transient-args 'justl-module-help-popup) (cons (just-module-name module) nil)))))
+
+(defun justl--go-to-module ()
+  "Go to the module path."
+  (interactive)
+  (let* ((module (justl--get-module-under-cursor)))
+    (find-file (just-module-source module))
+    (goto-char (point-min))))
+
+(defun justl--module-open-justl ()
+  "Open justl buffer for that specific module."
+  (interactive)
+  (let* ((module (justl--get-module-under-cursor))
+	 (default-directory (f-dirname (just-module-source module))))
+    (justl)))
+
 (provide 'justl)
 ;;; justl.el ends here
